/*
 * alignnetwork.cpp
 *
 * Copyright 2011-2012. All rights reserved.
 */

#include "alignnetwork.h"
#include "extern.h"

using namespace std;

extern const float LAMBDA;
extern const float ORTH_COEF;
extern const size_t NITERATION;
extern const float GRAPHLET_COEF;
extern const bool gVerbose;

int AlignNetwork(const char* hNodeFile,
                const char* gNodeFile,
                const char* hNetFile,
                const char* gNetFile,
                const char* sFile,
                const char* graphFile,
                const char* resultFile,
                const char* alignmentFile)
{
    Network hNetwork, gNetwork;

    hNetwork.InitializeNetwork(hNodeFile, hNetFile);
    gNetwork.InitializeNetwork(gNodeFile, gNetFile);

    cout << "Network 1 information:" << endl;
    hNetwork.PrintNetInfo();
    cout << "\nNetwork 2 information:" << endl;
    gNetwork.PrintNetInfo();

    SubNetwork hSubNetworks[hNetwork.nLinkedNode];
    SubNetwork gSubNetworks[gNetwork.nLinkedNode];

    hNetwork.GetSubNetworks(hSubNetworks);
    gNetwork.GetSubNetworks(gSubNetworks);

    SimilarityMatrix simMatrix[2];
    simMatrix[0].InitializeBinarySimMatrix(hNetwork, gNetwork, sFile);
    cout << "\nSimilarity information: " << endl;
    simMatrix[0].PrintSimilarityInfo();
    
    if (gVerbose == true)
    {
        cout << "\nSimilarity matrix : " << endl;
        simMatrix[0].PrintMatrix();
    }
    
    simMatrix[1].CopySimMatrix(simMatrix[0]);

    int opt = 0;

//  iterations
if(opt == 0)
{
    cout << "\nGenerating 1-neighborhood subnetwork similarity matrix..." << endl;
    size_t nIter = 0;
    int k = 0;
    while(nIter < NITERATION)
    {
        //cout << "Iteration " << nIter + 1 << endl;
        for (int i = 0; i < hNetwork.nNode; i++)
        {
            for (int j = 0; j < gNetwork.nNode; j++)
            {
                cout << "subnetwork " << i << " " << j << endl;
                if (hNetwork.isNonIsolated[i] == false || gNetwork.isNonIsolated[j] == false)
                {
                    if (hNetwork.isNonIsolated[i])
                    {
                        int hLinkedNode = hNetwork.GetLinkedNodeIndex(i);
                        IsoAlign(hNetwork, gNetwork, hSubNetworks[hLinkedNode], j, simMatrix[k], simMatrix[1-k]);
                    }
                    else if (gNetwork.isNonIsolated[j])
                    {
                        int gLinkedNode = gNetwork.GetLinkedNodeIndex(j);
                        IsoAlign(hNetwork, gNetwork, i, gSubNetworks[gLinkedNode], simMatrix[k], simMatrix[1-k]);
                    }
                    else
                    {
                        IsoAlign(hNetwork, gNetwork, i, j, simMatrix[k], simMatrix[1-k]);
                    }

                }
                else
                {
                    int hLinkedNode = hNetwork.GetLinkedNodeIndex(i);
                    int gLinkedNode = gNetwork.GetLinkedNodeIndex(j);
                    if (hNetwork.netIntNeighborVec[hLinkedNode].size() <= 10 && gNetwork.netIntNeighborVec[gLinkedNode].size() <= 10)
                    {
                        CplexSubILPAlign(hSubNetworks[hLinkedNode], gSubNetworks[gLinkedNode], simMatrix[k], simMatrix[1-k], true);
                    }
                    else
                    {
                        simMatrix[1-k].unalignedNodePairs[0].push_back(i);
                        simMatrix[1-k].unalignedNodePairs[1].push_back(j);
                        CplexSubILPAlign(hSubNetworks[hLinkedNode], gSubNetworks[gLinkedNode], simMatrix[k], simMatrix[1-k], false);
                    }
                }
                //usleep(50000);
            }
        }

        k = 1 - k;
        
        if (gVerbose == true)
        {
            cout << "similarity matrix: " << endl;
            simMatrix[k].PrintMatrix();
        }
        
        simMatrix[k].MatrixNormalization();
        
        if (gVerbose == true)
        {
            cout << endl;
            cout << "normalized matrix: " << endl;
            simMatrix[k].PrintMatrix();
        }
        nIter++;
    }
}

    float **simGraphletMatrix;
    float maxSimGraphlet = 0;
    float minSimGraphlet = 0;
    
    cout << "Generating graphlet similarity matrix..." << endl;
    simMatrix[0].GetSimGraphletMatrix(simGraphletMatrix, hNetwork, gNetwork, maxSimGraphlet, minSimGraphlet);
    
    if (gVerbose == true)
    {
        cout << "Graphlet similarity matrix :" << endl;
        simMatrix[0].PrintSimGraphletMatrix(simGraphletMatrix);
    }
    

    simMatrix[0].MatrixNormalization(simGraphletMatrix, maxSimGraphlet);
    if (gVerbose == true)
    {
        cout << endl;
        cout << "Normalized graphlet similarity matrix: " << endl;
        simMatrix[0].PrintSimGraphletMatrix(simGraphletMatrix);
        cout << endl;
    }
    
    int **finalSimMatrix;

    vector<string> vAssignAlignmentNodeStr[2];
    vector<string> vGlobalAlignmentNodeStr[2];

    finalSimMatrix = new int *[hNetwork.nNode];
    for (int i = 0; i < hNetwork.nNode; i++)
    {
        finalSimMatrix[i] = new int[gNetwork.nNode];
    }

    float maxAssignmentScore = 0;
    GetFinalSimMatrix(finalSimMatrix, simMatrix[0], simMatrix[1], simGraphletMatrix);
    
    if (gVerbose == true)
    {
        cout << "Final similarity matrix: " << endl;
        simMatrix[0].PrintSimGraphletMatrix(finalSimMatrix);
        cout << endl;
    }
    


    cout << "\n\nSolving the assignment problem: " << endl;
    CplexAssignmentILPAlign(hNetwork, gNetwork, finalSimMatrix, vAssignAlignmentNodeStr, maxAssignmentScore);
    if (gVerbose == true)
    {
        PrintAlignmentInfo(hNetwork, gNetwork, simMatrix[0], vAssignAlignmentNodeStr);
        cout << endl;
    }
    
    cout << "\n\nSolving the global ILP: " << endl;
    CplexGlobalILPAlign(hNetwork, gNetwork, simMatrix[0], finalSimMatrix, vGlobalAlignmentNodeStr, maxAssignmentScore);

    // Generate the graph
    GenerateAlignmentDotFile(hNetwork, gNetwork, graphFile, vGlobalAlignmentNodeStr, simMatrix[0]);
    GenerateAlignmentFile(hNetwork, gNetwork, alignmentFile, vGlobalAlignmentNodeStr, simMatrix[0]);
    PrintAlignmentInfo(hNetwork, gNetwork, simMatrix[0], vGlobalAlignmentNodeStr);
    simMatrix[0].CleanSimGraphletMatrix(simGraphletMatrix);
    simMatrix[0].CleanFinalSimMatrix(finalSimMatrix);


    int orthCount = 0;
    int loopCount = 0;
    for (int i = 0; i < vGlobalAlignmentNodeStr[0].size(); i++)
    {
        int node1 = hNetwork.GetNodeIndex(vGlobalAlignmentNodeStr[0][i]);
        int node2 = gNetwork.GetNodeIndex(vGlobalAlignmentNodeStr[1][i]);
        if (simMatrix[0].matrix[node1][node2] == 1)
        {
            orthCount += 1;
        }
        if (hNetwork.isLoop[node1] && gNetwork.isLoop[node2])
        {
            loopCount += 1;
        }
    }

    int nAlignEdge = 0;
    for (int i = 0; i < hNetwork.netIntVec[0].size(); i++)
    {
        for (int j = 0; j < gNetwork.netIntVec[0].size(); j++)
        {
            bool align00 = false;
            bool align01 = false;
            bool align10 = false;
            bool align11 = false;

            for (int k = 0; k < vGlobalAlignmentNodeStr[0].size(); k++)
            {
                int node0 = hNetwork.GetNodeIndex(vGlobalAlignmentNodeStr[0][k]);
                int node1 = gNetwork.GetNodeIndex(vGlobalAlignmentNodeStr[1][k]);
                if (hNetwork.netIntVec[0][i] == node0 && gNetwork.netIntVec[0][j] == node1)
                {
                    align00=true;
                }
                if (hNetwork.netIntVec[0][i] == node0 && gNetwork.netIntVec[1][j] == node1)
                {
                    align01=true;
                }
                if (hNetwork.netIntVec[1][i] == node0 && gNetwork.netIntVec[0][j] == node1)
                {
                    align10=true;
                }
                if (hNetwork.netIntVec[1][i] == node0 && gNetwork.netIntVec[1][j] == node1)
                {
                    align11=true;
                }
                if (align00==true && align11==true)
                {
                    nAlignEdge += 1;
                    break;
                }
                else if (align01==true && align10==true)
                {
                    nAlignEdge += 1;
                    break;
                }
            }
        }
    }

    ofstream rFile(resultFile);
    rFile << "lambda" << "\t"
          << "graphlet_coef" << "\t"
          << "orth_coef" << "\t"
          << "#nodes" << "\t"
          << "#edges" << "\t"
          << "#orths" << "\t"
          << "#loops" << endl;
    rFile << LAMBDA << "\t"
          << GRAPHLET_COEF << "\t"
          << ORTH_COEF << "\t"
          << vGlobalAlignmentNodeStr[0].size() << "\t"
          << nAlignEdge << "\t"
          << orthCount << "\t"
          << loopCount << endl;

    return 0;
}







