/*
 * cplexalign.cpp
 *
 * Copyright 2011-2012. All rights reserved.
 */


#include "cplexalign.h"

using namespace std;

extern const float ALPHA;
extern const float BETA;
extern const float LAMBDA;
extern const float GAMMA;
extern const float DELTA;
extern const char VARTYPE;
extern const bool gVerbose;

typedef IloArray<IloNumVarArray> NumVarMatrix;


int BuildGlobalModel(IloModel             mod,
                    NumVarMatrix        nodeVar,
                    NumVarMatrix        edgeHGVar,
                    Network&            hNetwork,
                    Network&            gNetwork,
                    SimilarityMatrix&   simMatrix,
                    int**             finalSimMatrix,
                    IloNumVar::Type     varType,
                    float               maxAssignmentScore)
{
    IloEnv env = mod.getEnv();

    int i, j, k;

    int nHNode = hNetwork.nNode;
    int nGNode = gNetwork.nNode;

    int nHEdge = hNetwork.nEdge;
    int nGEdge = gNetwork.nEdge;

    //define the variables
    for (i = 0; i < nHNode; i++)
    {
        nodeVar[i] = IloNumVarArray(env, nGNode, 0, 1, varType);
    }

    for (i = 0; i < nHEdge; i++)
    {
        edgeHGVar[i] = IloNumVarArray(env, nGEdge, 0, 1, varType);
    }

    //node restriction
    for (i = 0; i < nHNode; i++)
    {
        IloExpr expr(env);
        for (j = 0; j< nGNode; j++)
        {
            expr += nodeVar[i][j];
        }
        mod.add( expr <= 1);
        expr.end();
    }

    for (i = 0; i < nGNode; i++)
    {
        IloExpr expr(env);
        for (j = 0; j < nHNode; j++)
        {
            expr += nodeVar[j][i];
        }
        mod.add(expr <= 1);
        expr.end();
    }


    //assignment resitriction
    IloExpr expr(env);
    for (i = 0; i < nHNode; i++)
    {
        for (j = 0; j < nGNode; j++)
        {
            expr += finalSimMatrix[i][j] * nodeVar[i][j];
        }
    }
    mod.add(expr == maxAssignmentScore);
    expr.end();

    //edge restriction

    for (i = 0; i < nHEdge; i++)
    {
        int hNode1 = hNetwork.netIntVec[0][i];
        int hNode2 = hNetwork.netIntVec[1][i];
        for (j = 0; j < nGEdge; j++)
        {
            int gNode1 = gNetwork.netIntVec[0][j];
            int gNode2 = gNetwork.netIntVec[1][j];

            mod.add(edgeHGVar[i][j] - (nodeVar[hNode1][gNode1] + nodeVar[hNode1][gNode2]) <= 0);
            mod.add(edgeHGVar[i][j] - (nodeVar[hNode2][gNode1] + nodeVar[hNode2][gNode2]) <= 0);
        }
    }

    IloExpr nodeObj(env);
    IloExpr edgeObj(env);

    for (i = 0; i < nHNode; i++)
    {
        for (j = 0; j < nGNode; j++)
        {
            nodeObj += ((GAMMA - DELTA)*simMatrix.matrix[i][j] + DELTA) * nodeVar[i][j];
            edgeObj += hNetwork.isLoop[i]*gNetwork.isLoop[j]*(ALPHA - 2*BETA) * nodeVar[i][j];
        }
    }

    for (i = 0; i < nHEdge; i++)
    {
        for (j = 0; j < nGEdge; j++)
        {
            edgeObj += (ALPHA - 2*BETA) * edgeHGVar[i][j]; //ALPHA*SUM(edgeHGVar[i][j]) + BETA*(nHEdge + nGEdge - 2*SUM(edgeHGVar[i][j]))
        }
    }

    edgeObj += BETA*(nHEdge + nGEdge + hNetwork.nLoop + gNetwork.nLoop);

    mod.add(IloMaximize(env, LAMBDA*nodeObj + (1-LAMBDA)*edgeObj));

    nodeObj.end();
    edgeObj.end();

    return 0;
}







int BuildGlobalModel(IloModel             mod,
                    NumVarMatrix        nodeVar,
                    NumVarMatrix        edgeHGVar,
                    Network&            hNetwork,
                    Network&            gNetwork,
                    SimilarityMatrix&   simMatrix,
                    vector<int>*        fixedVarIntList,
                    IloNumVar::Type     varType)
{
    IloEnv env = mod.getEnv();

    int i, j, k;

    int nHNode = hNetwork.nNode;
    int nGNode = gNetwork.nNode;

    int nHEdge = hNetwork.nEdge;
    int nGEdge = gNetwork.nEdge;

    //define the variables
    for (i = 0; i < nHNode; i++)
    {
        nodeVar[i] = IloNumVarArray(env, nGNode, 0, 1, varType);
    }

    for (i = 0; i < nHEdge; i++)
    {
        edgeHGVar[i] = IloNumVarArray(env, nGEdge, 0, 1, varType);
    }

    //node restriction
    for (i = 0; i < nHNode; i++)
    {
        IloExpr expr(env);
        for (j = 0; j< nGNode; j++)
        {
            expr += nodeVar[i][j];
        }
        mod.add( expr <= 1);
        expr.end();
    }

    for (i = 0; i < nGNode; i++)
    {
        IloExpr expr(env);
        for (j = 0; j < nHNode; j++)
        {
            expr += nodeVar[j][i];
        }
        mod.add(expr <= 1);
        expr.end();
    }

    IloExpr nodeObj(env);

    for (i = 0; i < nHNode; i++)
    {
        for (j = 0; j < nGNode; j++)
        {
            nodeObj += ((GAMMA - DELTA)*simMatrix.matrix[i][j] + DELTA) * nodeVar[i][j];
        }
    }


    IloExpr edgeObj(env);

    for (i = 0; i < nHEdge; i++)
    {
        for (j = 0; j < nGEdge; j++)
        {
            edgeObj += (ALPHA - 2*BETA) * edgeHGVar[i][j]; //ALPHA*SUM(edgeHGVar[i][j]) + BETA*(nHEdge + nGEdge - 2*SUM(edgeHGVar[i][j]))
        }
    }

    edgeObj += BETA*(nHEdge + nGEdge);

    mod.add(IloMaximize(env, LAMBDA*nodeObj + (1-LAMBDA)*edgeObj));

    nodeObj.end();
    edgeObj.end();

    return 0;
}

int BuildAssignmentModel(IloModel             mod,
                        NumVarMatrix        nodeVar,
                        Network&            hNetwork,
                        Network&            gNetwork,
                        int**               simMatrix,
                        IloNumVar::Type     varType)
{
    IloEnv env = mod.getEnv();

    int i, j, k;

    int nHNode = hNetwork.nNode;
    int nGNode = gNetwork.nNode;

    for (i = 0; i < nHNode; i++)
    {
        nodeVar[i] = IloNumVarArray(env, nGNode, 0, 1, varType);
    }

    //node restriction
    for (i = 0; i < nHNode; i++)
    {
        IloExpr expr(env);
        for (j = 0; j< nGNode; j++)
        {
            expr += nodeVar[i][j];
        }
        mod.add( expr <= 1);
        expr.end();
    }

    for (i = 0; i < nGNode; i++)
    {
        IloExpr expr(env);
        for (j = 0; j < nHNode; j++)
        {
            expr += nodeVar[j][i];
        }
        mod.add(expr <= 1);
        expr.end();
    }

    IloExpr nodeObj(env);

    for (i = 0; i < nHNode; i++)
    {
        for (j = 0; j < nGNode; j++)
        {
            nodeObj += simMatrix[i][j] * nodeVar[i][j];
        }
    }

    mod.add(IloMaximize(env, nodeObj));

    nodeObj.end();

    return 0;
}



int buildAlignSubNetworkModel(IloModel              mod,
                                NumVarMatrix        nodeVar,
                                NumVarMatrix        edgeHGVar,
                                SubNetwork&         hSubNetwork,
                                SubNetwork&         gSubNetwork,
                                SimilarityMatrix&   simMatrix,
                                IloNumVar::Type     varType)
{
    IloEnv env = mod.getEnv();

    int i, j, k;

    int nHNode = hSubNetwork.nNode;
    int nGNode = gSubNetwork.nNode;

    int nHNeiNode = nHNode - 1;
    int nGNeiNode = nGNode - 1;

    //int nNodeVar = nHNeiNode + nGNeiNode;

    int nHEdge = hSubNetwork.nEdge;
    int nGEdge = gSubNetwork.nEdge;

    int nHNtrvlEdge = nHEdge - nHNeiNode; // Non-trival Edge
    int nGNtrvlEdge = nGEdge - nGNeiNode;


    int nHPNtrvlEdge = (nHNeiNode*nHNeiNode - nHNeiNode)/2 - nHNtrvlEdge; //Potential Non-trival Edge
    int nGPNtrvlEdge = (nGNeiNode*nGNeiNode - nGNeiNode)/2 - nGNtrvlEdge;

    //define the variables
    for (i = 0; i < nHNeiNode; i++)
    {
        nodeVar[i] = IloNumVarArray(env, nGNeiNode, 0, 1, varType);
    }

    for (i = 0; i < nHNtrvlEdge; i++)
    {
        edgeHGVar[i] = IloNumVarArray(env, nGNtrvlEdge, 0, 1, varType);
    }
    
    //node restriction
    for (i = 0; i < nHNeiNode; i++)
    {
        IloExpr expr(env);
        for (j = 0; j< nGNeiNode; j++)
        {
            expr += nodeVar[i][j];
        }
        mod.add( expr <= 1);
        expr.end();
    }

    for (i = 0; i < nGNeiNode; i++)
    {
        IloExpr expr(env);
        for (j = 0; j < nHNeiNode; j++)
        {
            expr += nodeVar[j][i];
        }
        mod.add(expr <= 1);
        expr.end();
    }

    //edge restriction

    for (i = 0; i < nHNtrvlEdge; i++)
    {
        int hNode1 = hSubNetwork.netIntVec[0][i];
        int hNode2 = hSubNetwork.netIntVec[1][i];
        for (j = 0; j < nGNtrvlEdge; j++)
        {
            int gNode1 = gSubNetwork.netIntVec[0][j];
            int gNode2 = gSubNetwork.netIntVec[1][j];

            mod.add(edgeHGVar[i][j] - (nodeVar[hNode1][gNode1] + nodeVar[hNode1][gNode2]) <= 0);
            mod.add(edgeHGVar[i][j] - (nodeVar[hNode2][gNode1] + nodeVar[hNode2][gNode2]) <= 0);
        }
    }

    IloExpr nodeObj(env);

    int hIndex, gIndex;

    hIndex = hSubNetwork.centerNode;
    gIndex = gSubNetwork.centerNode;
    nodeObj += ((GAMMA - DELTA)*simMatrix.matrix[hIndex][gIndex] + DELTA);

    for (i = 0; i < nHNeiNode; i++)
    {
        for (j = 0; j < nGNeiNode; j++)
        {
            hIndex = hSubNetwork.GetNetIndex(i);
            gIndex = gSubNetwork.GetNetIndex(j);
            nodeObj += ((GAMMA - DELTA) * simMatrix.matrix[hIndex][gIndex] + DELTA) * nodeVar[i][j];
        }
    }


    IloExpr edgeObj(env);

    for (i = 0; i < nHNtrvlEdge; i++)
    {
        for (j = 0; j < nGNtrvlEdge; j++)
        {
            edgeObj += (ALPHA - 2*BETA) * edgeHGVar[i][j];
        }
    }

    for (i = 0; i < nHNeiNode; i++)
    {
        for (j = 0; j < nGNeiNode; j++)
        {
            edgeObj += (ALPHA - 2*BETA + hSubNetwork.isLoop[i]*gSubNetwork.isLoop[j]*(ALPHA - 2*BETA)) * nodeVar[i][j];
        }
    }

    edgeObj += BETA*(nHNeiNode + nGNeiNode + nHNtrvlEdge + nGNtrvlEdge + hSubNetwork.nLoop + gSubNetwork.nLoop);

    if (hSubNetwork.isCenterNodeLoop && gSubNetwork.isCenterNodeLoop)
    {
        edgeObj += ALPHA;
    }
    else if (hSubNetwork.isCenterNodeLoop || gSubNetwork.isCenterNodeLoop)
    {
        edgeObj += BETA;
    }



    mod.add(IloMaximize(env, LAMBDA*nodeObj + (1-LAMBDA)*edgeObj));

    nodeObj.end();
    edgeObj.end();

    return 0;
}


int buildAlignSubNetworkModel(IloModel              mod,
                                NumVarMatrix        nodeVar,
                                SubNetwork&         hSubNetwork,
                                SubNetwork&         gSubNetwork,
                                SimilarityMatrix&   simMatrix,
                                IloNumVar::Type     varType)
{
    IloEnv env = mod.getEnv();

    int i, j, k;



    int nHNode = hSubNetwork.nNode;
    int nGNode = gSubNetwork.nNode;

    int nHNeiNode = nHNode - 1;
    int nGNeiNode = nGNode - 1;

    int nHEdge = hSubNetwork.nEdge;
    int nGEdge = gSubNetwork.nEdge;

    int nHNtrvlEdge = nHEdge - nHNeiNode; // Non-trival Edge
    int nGNtrvlEdge = nGEdge - nGNeiNode;


    int nHPNtrvlEdge = (nHNeiNode*nHNeiNode - nHNeiNode)/2 - nHNtrvlEdge; //Potential Non-trival Edge
    int nGPNtrvlEdge = (nGNeiNode*nGNeiNode - nGNeiNode)/2 - nGNtrvlEdge;

    //define the variables
    for (i = 0; i < nHNeiNode; i++)
    {
        nodeVar[i] = IloNumVarArray(env, nGNeiNode, 0, 1, varType);
    }


    //node restriction
    for (i = 0; i < nHNeiNode; i++)
    {
        IloExpr expr(env);
        for (j = 0; j< nGNeiNode; j++)
        {
            expr += nodeVar[i][j];
        }
        mod.add( expr <= 1);
        expr.end();
    }

    for (i = 0; i < nGNeiNode; i++)
    {
        IloExpr expr(env);
        for (j = 0; j < nHNeiNode; j++)
        {
            expr += nodeVar[j][i];
        }
        mod.add(expr <= 1);
        expr.end();
    }

    IloExpr nodeObj(env);

    int hIndex, gIndex;

    hIndex = hSubNetwork.centerNode;
    gIndex = gSubNetwork.centerNode;
    nodeObj += ((GAMMA - DELTA)*simMatrix.matrix[hIndex][gIndex] + DELTA);

    for (i = 0; i < nHNeiNode; i++)
    {
        for (j = 0; j < nGNeiNode; j++)
        {
            hIndex = hSubNetwork.GetNetIndex(i);
            gIndex = gSubNetwork.GetNetIndex(j);
            nodeObj += ((GAMMA - DELTA) * simMatrix.matrix[hIndex][gIndex] + DELTA) * nodeVar[i][j];
        }
    }


    IloExpr edgeObj(env);

    int maxNEdgeAligned = nHNtrvlEdge < nGNtrvlEdge ? nHNtrvlEdge : nGNtrvlEdge;

    edgeObj += (ALPHA - 2*BETA) * maxNEdgeAligned;

    for (i = 0; i < nHNeiNode; i++)
    {
        for (j = 0; j < nGNeiNode; j++)
        {
            edgeObj += (ALPHA - 2*BETA + hSubNetwork.isLoop[i]*gSubNetwork.isLoop[j]*(ALPHA - 2*BETA)) * nodeVar[i][j];
        }
    }

    edgeObj += BETA*(nHNeiNode + nGNeiNode + nHNtrvlEdge + nGNtrvlEdge + hSubNetwork.nLoop + gSubNetwork.nLoop);

    if (hSubNetwork.isCenterNodeLoop && gSubNetwork.isCenterNodeLoop)
    {
        edgeObj += ALPHA;
    }
    else if (hSubNetwork.isCenterNodeLoop || gSubNetwork.isCenterNodeLoop)
    {
        edgeObj += BETA;
    }

    mod.add(IloMaximize(env, LAMBDA*nodeObj + (1-LAMBDA)*edgeObj));

    nodeObj.end();
    edgeObj.end();

    return 0;
}

int CplexAssignmentILPAlign(Network&        hNetwork,
                            Network&        gNetwork,
                            int**           simMatrix,
                            vector<string>* vAlignmentNodeStr,
                            float&          maxAssignmentScore)
{
    IloEnv env;

    try {


        timeb startTime, endTime;
        ftime(&startTime);

        IloNumVar::Type varType = ILOINT;
        if (VARTYPE == 'F') varType = ILOFLOAT;

        IloModel mod(env);

        NumVarMatrix nodeVar(env, hNetwork.nNode);

        BuildAssignmentModel(mod,
                            nodeVar,
                            hNetwork,
                            gNetwork,
                            simMatrix,
                            varType);

        IloCplex cplex(mod);

        cplex.solve();
        cplex.out() << "solution status = " << cplex.getStatus() << endl;
        cplex.out() << "Score    = " << cplex.getObjValue() << endl;
        //cout << "alignment: " << endl;

        int nAlignNode = 0;
        int nAlignEdge = 0;
        int nAlignOrth = 0;
        for (int i = 0; i < hNetwork.nNode; i++)
        {
            for (int j = 0; j < gNetwork.nNode; j++)
            {
                if (cplex.getValue(nodeVar[i][j]) > 0.5)
                {
                    //cout << hNetwork.nodeStrVec[i] << " " << gNetwork.nodeStrVec[j] << " " << simMatrix[i][j] << endl;
                    vAlignmentNodeStr[0].push_back(hNetwork.nodeStrVec[i]);
                    vAlignmentNodeStr[1].push_back(gNetwork.nodeStrVec[j]);
                    nAlignNode += 1;
                }
            }
        }

        maxAssignmentScore = cplex.getObjValue();

        ftime(&endTime);
    }



    catch (IloException& ex) {
        cerr << "Error: " << ex << endl;
    }
    catch (...) {
        cerr << "Error" << endl;
    }

    env.end();
    return 0;
}

int CplexAssignmentILPAlign(Network&        hNetwork,
                            Network&        gNetwork,
                            int**           simMatrix,
                            vector<string>* vAlignmentNodeStr)
{
    IloEnv env;

    try {


        timeb startTime, endTime;
        ftime(&startTime);

        IloNumVar::Type varType = ILOINT;
        if (VARTYPE == 'F') varType = ILOFLOAT;

        IloModel mod(env);

        NumVarMatrix nodeVar(env, hNetwork.nNode);

        BuildAssignmentModel(mod,
                            nodeVar,
                            hNetwork,
                            gNetwork,
                            simMatrix,
                            varType);

        IloCplex cplex(mod);

        cplex.solve();
        cplex.out() << "solution status = " << cplex.getStatus() << endl;
        cplex.out() << "Score    = " << cplex.getObjValue() << endl;
        cout << "alignment: " << endl;

        int nAlignment = 0;
        for (int i = 0; i < hNetwork.nNode; i++)
        {
            for (int j = 0; j < gNetwork.nNode; j++)
            {
                if (cplex.getValue(nodeVar[i][j]) == 1)
                {
                    cout << hNetwork.nodeStrVec[i] << " " << gNetwork.nodeStrVec[j];
                    printf(" %.6f\n", (double)simMatrix[i][j]);
                    vAlignmentNodeStr[0].push_back(hNetwork.nodeStrVec[i]);
                    vAlignmentNodeStr[1].push_back(gNetwork.nodeStrVec[j]);
                    nAlignment += 1;
                }
            }
        }

        cout << "nAlignment : " << nAlignment << endl;


        ftime(&endTime);
    }



    catch (IloException& ex) {
        cerr << "Error: " << ex << endl;
    }
    catch (...) {
        cerr << "Error" << endl;
    }

    env.end();
    return 0;
}

int CplexGlobalILPAlign(Network&          hNetwork,
                        Network&          gNetwork,
                        SimilarityMatrix& simMatrix,
                        int**             finalSimMatrix,
                        vector<string>*   vAlignmentNodeStr,
                        float             maxAssignmentScore)
{
    IloEnv env;

    try {


        timeb startTime, endTime;
        ftime(&startTime);

        IloNumVar::Type varType = ILOINT;
        if (VARTYPE == 'F') varType = ILOFLOAT;

        IloModel mod(env);

        NumVarMatrix nodeVar(env, hNetwork.nNode);
        NumVarMatrix edgeHGVar(env, hNetwork.nEdge);

        BuildGlobalModel(mod,
                        nodeVar,
                        edgeHGVar,
                        hNetwork,
                        gNetwork,
                        simMatrix,
                        finalSimMatrix,
                        varType,
                        maxAssignmentScore);



        IloCplex cplex(mod);
        //cplex.setParam(IloCplex::TiLim, 7200);
        
        cplex.solve();
        cplex.out() << "solution status = " << cplex.getStatus() << endl;
        cplex.out() << "Score    = " << cplex.getObjValue() << endl;

        for (int i = 0; i < hNetwork.nNode; i++)
        {
            for (int j = 0; j < gNetwork.nNode; j++)
            {
                if (cplex.getValue(nodeVar[i][j]) > 0.5)
                {
                    vAlignmentNodeStr[0].push_back(hNetwork.nodeStrVec[i]);
                    vAlignmentNodeStr[1].push_back(gNetwork.nodeStrVec[j]);
                }
            }
        }

        ftime(&endTime);
    }



    catch (IloException& ex) {
        cerr << "Error: " << ex << endl;
    }
    catch (...) {
        cerr << "Error" << endl;
    }

    env.end();
    return 0;
}


int CplexGlobalILPAlign(Network&          hNetwork,
                        Network&          gNetwork,
                        SimilarityMatrix& simMatrix,
                        vector<string>*   vFixedNodeStr)
{
    IloEnv env;

    try {


        timeb startTime, endTime;
        ftime(&startTime);
        
        IloNumVar::Type varType = ILOINT;
        if (VARTYPE == 'F') varType = ILOFLOAT;

        IloModel mod(env);

        NumVarMatrix nodeVar(env, hNetwork.nNode);
        NumVarMatrix edgeHGVar(env, hNetwork.nEdge);

        vector<int> fixedVarIntList[2];

        for (int i = 0; i < vFixedNodeStr[0].size(); i++)
        {
            fixedVarIntList[0].push_back(hNetwork.GetNodeIndex(vFixedNodeStr[0][i]));
            fixedVarIntList[1].push_back(gNetwork.GetNodeIndex(vFixedNodeStr[1][i]));
        }

        BuildGlobalModel(mod,
                    nodeVar,
                    edgeHGVar,
                    hNetwork,
                    gNetwork,
                    simMatrix,
                    fixedVarIntList,
                    varType);



        IloCplex cplex(mod);

        cplex.solve();

        cplex.out() << "solution status = " << cplex.getStatus() << endl;
        cplex.out() << "Score    = " << cplex.getObjValue() << endl;

        cout << "alignment: " << endl;
        int nAlignNode = 0;
        int nAlignEdge = 0;

        for (int i = 0; i < hNetwork.nNode; i++)
        {
            for (int j = 0; j < gNetwork.nNode; j++)
            {
                if (cplex.getValue(nodeVar[i][j]) == 1)
                {
                    cout << hNetwork.nodeStrVec[i] << " " << gNetwork.nodeStrVec[j];
                    printf(" %.6f\n", simMatrix.matrix[i][j]);
                    nAlignNode += 1;
                }
            }
        }

        for (int i = 0; i < hNetwork.nEdge; i++)
        {
            for (int j = 0; j < gNetwork.nEdge; j++)
            {
                if (cplex.getValue(edgeHGVar[i][j]) == 1)
                {
                    nAlignEdge += 1;
                }
            }
        }

        cout << "No. of aligned nodes : " << nAlignNode << endl;
        cout << "No. of aligned edges : " << nAlignEdge << endl;

        ftime(&endTime);
    }



    catch (IloException& ex) {
        cerr << "Error: " << ex << endl;
    }
    catch (...) {
        cerr << "Error" << endl;
    }

    env.end();
    return 0;
}


int CplexSubILPAlign(SubNetwork&       hSubNetwork,
                     SubNetwork&       gSubNetwork,
                     SimilarityMatrix& inSimMatrix,
                     SimilarityMatrix& outSimMatrix,
                     bool              lowDegree)
{
    IloEnv env;

    try {
        timeb startTime, endTime;
        ftime(&startTime);

        IloNumVar::Type varType = ILOINT;
        if (VARTYPE == 'F') varType = ILOFLOAT;

        IloModel mod(env);

        NumVarMatrix nodeVar(env, hSubNetwork.nNode - 1);
        NumVarMatrix edgeHGVar(env, hSubNetwork.nEdge - hSubNetwork.nNode + 1);
        if (lowDegree == true)
        {
            buildAlignSubNetworkModel(mod,
                                    nodeVar,
                                    edgeHGVar,
                                    hSubNetwork,
                                    gSubNetwork,
                                    inSimMatrix,
                                    varType);
        }
        else
        {
            buildAlignSubNetworkModel(mod,
                                    nodeVar,
                                    hSubNetwork,
                                    gSubNetwork,
                                    inSimMatrix,
                                    varType);
        }



        IloCplex cplex(mod);
        if(gVerbose == false)
        {
            cplex.setOut(env.getNullStream());
        }
        cplex.solve();
        
        cplex.out() << "solution status = " << cplex.getStatus() << endl;
        cplex.out() << "Score    = " << cplex.getObjValue() << endl;

        int hFixedNode = hSubNetwork.centerNode;
        int gFixedNode = gSubNetwork.centerNode;
        outSimMatrix.matrix[hFixedNode][gFixedNode] = cplex.getObjValue();
        outSimMatrix.UpdateMaxMinSim(outSimMatrix.matrix[hFixedNode][gFixedNode]);

        ftime(&endTime);
    }



    catch (IloException& ex) {
        cerr << "Error: " << ex << endl;
    }
    catch (...) {
        cerr << "Error" << endl;
    }

    env.end();
    return 0;
}


int IsoAlign(Network&          hNetwork,
             Network&          gNetwork,
             int               hNode,
             int               gNode,
             SimilarityMatrix& inSimMatrix,
             SimilarityMatrix& outSimMatrix)
{
    float edgeScore = 0;
    float nodeScore = (GAMMA - DELTA) * inSimMatrix.matrix[hNode][gNode] + DELTA;

    if (hNetwork.isLoop[hNode] && gNetwork.isLoop[gNode])
    {
        edgeScore += ALPHA;
    }
    else if (hNetwork.isLoop[hNode] || gNetwork.isLoop[gNode])
    {
        edgeScore += BETA;
    }

    outSimMatrix.matrix[hNode][gNode] = (1-LAMBDA)*edgeScore + LAMBDA*nodeScore;
    outSimMatrix.UpdateMaxMinSim(outSimMatrix.matrix[hNode][gNode]);

    return 0;
}

int IsoAlign(Network&          hNetwork,
             Network&          gNetwork,
             SubNetwork&       hSubnetwork,
             int               gNode,
             SimilarityMatrix& inSimMatrix,
             SimilarityMatrix& outSimMatrix)
{
    int hNode = hSubnetwork.centerNode;

    float edgeScore = 0;
    float nodeScore = (GAMMA - DELTA) * inSimMatrix.matrix[hNode][gNode] + DELTA;

    edgeScore += hSubnetwork.nEdge*BETA;

    if (hNetwork.isLoop[hNode] && gNetwork.isLoop[gNode])
    {
        edgeScore += ALPHA;
    }
    else if (hNetwork.isLoop[hNode] || gNetwork.isLoop[gNode])
    {
        edgeScore += BETA;
    }

    edgeScore += hSubnetwork.nLoop*BETA;

    outSimMatrix.matrix[hNode][gNode] = (1-LAMBDA)*edgeScore + LAMBDA*nodeScore;
    outSimMatrix.UpdateMaxMinSim(outSimMatrix.matrix[hNode][gNode]);
    return 0;
}

int IsoAlign(Network&          hNetwork,
             Network&          gNetwork,
             int               hNode,
             SubNetwork&       gSubnetwork,
             SimilarityMatrix& inSimMatrix,
             SimilarityMatrix& outSimMatrix)
{

    int gNode = gSubnetwork.centerNode;

    float edgeScore = 0;
    float nodeScore = (GAMMA - DELTA) * inSimMatrix.matrix[hNode][gNode] + DELTA;

    edgeScore += gSubnetwork.nEdge*BETA; 


    if (hNetwork.isLoop[hNode] && gNetwork.isLoop[gNode])
    {
        edgeScore += ALPHA;
    }
    else if (hNetwork.isLoop[hNode] || gNetwork.isLoop[gNode])
    {
        edgeScore += BETA;
    }

    edgeScore += gSubnetwork.nLoop*BETA;

    outSimMatrix.matrix[hNode][gNode] = (1-LAMBDA)*edgeScore + LAMBDA*nodeScore;
    outSimMatrix.UpdateMaxMinSim(outSimMatrix.matrix[hNode][gNode]);

    return 0;
}
