/*
 * main.cpp
 *
 * Copyright 2011-2012. All rights reserved.
 */

#include <sys/timeb.h>
#include <iostream>
#include <fstream>
#include <sstream>
#include <getopt.h>
#include "alignnetwork.h"

using namespace std;

/*
 * Constants
 */
const char FILE_SEP('/'); // UNIX-specific

/*
 * Global Variables
 */
bool gVerbose = false;
size_t NITERATION = 1;
char VARTYPE = 'I';


float GRAPHLET_COEF = 0.1; // weight of graphlet similarity matrix
float ORTH_COEF     = 0.8; // weight of ortholog matrix

float LAMBDA        = 0.5; // weight of node score

// edge scores
float ALPHA = 1.6;  // edge matching score
float BETA  = -0.3; // else

// node scores
float GAMMA = 4.4;  // node ortholog score
float DELTA = -1.6; // else

/*
 * Functions
 */

inline bool isAbsolute(const std::string& filename)
 {
    return (filename[0] == FILE_SEP);
 }
 
int Usage (const char* const binary,
           std::ostream& os = std::cerr)
{
    os << "Usage: " << binary << "\n"
       << " [-m|--net1-file] [file name]\n"
       << " [-n|--net2-file] [file name]\n"
       << " [-i|--node1-file] [file name]\n"
       << " [-j|--node2-file] [file name]\n"
       << " [-s|--orth-file] [file name]\n"
       << " [-o|--output-dir] [file name]\n"
       << " [-r|--result-file] [file name]\n"
       << " [-k|--graph-file] [file name]\n"
       << " [-p|--alignment-file] [file name]\n"
       << " [-l|--lambda] [0-1]\n"
       << " [-g|--graphlet-coef] [0-1]\n"
       << " [-f|--orth-coef] [0-1]\n"
       << " [-a|--alpha]\n"
       << " [-b|--beta]\n"
       << " [-c|--gamma]\n"
       << " [-d|--delta]\n"
       << " [-h|--help]\n"
       << " [-v|--verbose]\n"
       << endl;
    return -1;
}


int main(int argc, char** argv)
{
    string hNodeFile;
    string gNodeFile;
    string hNetFile;
    string gNetFile;
    string outputDir(".");
    string graphFile = "alignment_graph.dot";
    string alignmentFile = "alignment.tsv";
    string resultFile = "result.tsv";
    string sFile;

    int opt;

    while (1)
    {
        static struct option long_options[]=
        {
            {"help",            no_argument,       0, 'h'},
            {"net1-file",       required_argument, 0, 'm'},
            {"net2-file",       required_argument, 0, 'n'},
            {"node1-file",      required_argument, 0, 'i'},
            {"node2-file",      required_argument, 0, 'j'},
            {"orth-file",       required_argument, 0, 's'},
            {"output-dir",      required_argument, 0, 'o'},
            {"result-file",     required_argument, 0, 'r'},
            {"graph-file",      required_argument, 0, 'k'},
            {"alignment-file",  required_argument, 0, 'p'},
            {"lambda",          required_argument, 0, 'l'},
            {"graphlet-coef",   required_argument, 0, 'g'},
            {"orth-coef",       required_argument, 0, 'f'},
            {"alpha",           required_argument, 0, 'a'},
            {"beta",            required_argument, 0, 'b'},
            {"gamma",           required_argument, 0, 'c'},
            {"delta",           required_argument, 0, 'd'},
            {"verbose",         no_argument,       0, 'v'},
            {0,                 0,                 0, 0}
        };

        opt = getopt_long (argc, argv, ":hm:n:i:j:s:o:r:k:p:l:g:f:a:b:c:d:v:", long_options, NULL);

        if (opt == -1)
            break;

        stringstream ssTmp;

        switch(opt)
        {
            case 'm':
                hNetFile = optarg;
                break;
            case 'n':
                gNetFile = optarg;
                break;
            case 'i':
                hNodeFile = optarg;
                break;
            case 'j':
                gNodeFile = optarg;
                break;
            case 's':
                sFile = optarg;
                break;
            case 'o':
                outputDir = optarg;
                break;
            case 'r':
                resultFile = optarg;
                break;
            case 'k':
                graphFile = optarg;
                break;
            case 'p':
                alignmentFile = optarg;
                break;
            case 'l':
                ssTmp << optarg;
                ssTmp >> LAMBDA;
                break;
            case 'a':
                ssTmp << optarg;
                ssTmp >> ALPHA;
                break;
            case 'b':
                ssTmp << optarg;
                ssTmp >> BETA;
                break;
            case 'c':
                ssTmp << optarg;
                ssTmp >> GAMMA;
                break;
            case 'd':
                ssTmp << optarg;
                ssTmp >> DELTA;
                break;
            case 'g':
                ssTmp << optarg;
                ssTmp >> GRAPHLET_COEF;
                break;
            case 'f':
                ssTmp << optarg;
                ssTmp >> ORTH_COEF;
                break;
            case 'v':
                gVerbose = true;
                break;
            case 'h':
                Usage(argv[0]);
                exit(EXIT_SUCCESS);
                /*NOTREACHED*/
            case ':':
                cerr << "Missing argument!" << endl;
                Usage(argv[0]);
                exit(EXIT_FAILURE);
                /*NOTREACHED*/
            case '?':
                cerr << "Unknown argument!" << endl;
                Usage(argv[0]);
                /*FALLTHROUGH*/
            default:
                exit(EXIT_FAILURE);
                /*NOTREACHED*/
        }
    }

    if (!isAbsolute(resultFile))
    {
        resultFile = outputDir + FILE_SEP + resultFile;
    }
    if (!isAbsolute(graphFile))
    {
        graphFile = outputDir + FILE_SEP + graphFile;
    }
    if (!isAbsolute(alignmentFile))
    {
        alignmentFile = outputDir + FILE_SEP + alignmentFile;
    }
    
    
    ifstream hNDP, gNDP, hNFP, gNFP, sFP;

    hNDP.open(hNodeFile.c_str());
    gNDP.open(gNodeFile.c_str());
    hNFP.open(hNetFile.c_str());
    gNFP.open(gNetFile.c_str());
    sFP.open(sFile.c_str());

    if (!hNDP || !gNDP || !hNFP || !gNFP || !sFP)
    {
        cerr << "Input file error!" << endl;
        return -1;
    }
    hNDP.close();
    gNDP.close();
    hNFP.close();
    gNFP.close();
    sFP.close();

    timeb startTime, endTime;
    ftime(&startTime);

    AlignNetwork(hNodeFile.c_str(),
                 gNodeFile.c_str(),
                 hNetFile.c_str(),
                 gNetFile.c_str(),
                 sFile.c_str(),
                 graphFile.c_str(),
                 resultFile.c_str(),
                 alignmentFile.c_str());



    ftime(&endTime);
    cout << "Total elapsed time is " << (float)(endTime.time - startTime.time) + (float)((endTime.millitm - startTime.millitm)/1000.) << "s" << endl;

    return 0;
}


