/*
 * similaritymatrix.cpp
 *
 * Copyright 2011-2012. All rights reserved.
 */

#include <cstdlib>
#include "similaritymatrix.h"

using namespace std;

extern const float ALPHA;
extern const float GAMMA;

const float ORTH_SCORE = GAMMA;
const float EDGE_SCORE = ALPHA;
const float NODE_WEIGHT = 0.5;
const float EDGE_WEIGHT = 1-NODE_WEIGHT;

SimilarityMatrix::SimilarityMatrix() :
    nHNode(0),
    nGNode(0),
    nSim(0),
    maxSim(0),
    minSim(0),
    matrix(NULL)
{
}

SimilarityMatrix::~SimilarityMatrix()
{
    for(int i = 0; i < nHNode; i++)
    {
        delete[] matrix[i];
        matrix[i] = NULL;
    }
    delete[] matrix;
    matrix = NULL;
}

int SimilarityMatrix::InitializeSimMatrix(Network& hNetwork, Network& gNetwork, const char* sFile)
{
    nHNode = hNetwork.nNode;
    nGNode = gNetwork.nNode;

    ifstream ifs(sFile);

    if (!ifs.good())
    {
        cerr << "Input file error!" << endl;
        exit(-1);
    }

    string hStrNode, gStrNode;
    float similarity;
    while(ifs >> hStrNode >> gStrNode >> similarity)
    {
        if (similarity > 0)
        {
            simStrList[0].push_back(hStrNode);
            simStrList[1].push_back(gStrNode);
            similarityVec.push_back(similarity);
        }
        hStrNode.clear();
        gStrNode.clear();
    }
    nSim = similarityVec.size();

    if (nSim > nHNode*nGNode)
    {
        cerr << "the scale of the similarity matrix from file is not as the same as nHNode*nGNode" << endl;
        exit(0);
    }
    matrix = new float* [nHNode];

    for (int i = 0; i < nHNode; i++)
    {
        matrix[i] = new float[nGNode];
        for (int j = 0; j < nGNode; j++)
        {
            matrix[i][j] = 0;
        }
    }

    int hNode, gNode;
    for (int i = 0; i < nSim; i++)
    {
        hNode = hNetwork.GetNodeIndex(simStrList[0][i]);
        gNode = gNetwork.GetNodeIndex(simStrList[1][i]);
        matrix[hNode][gNode] = similarityVec[i];
        if (i == 0)
        {
            maxSim = similarityVec[i];
            minSim = similarityVec[i];
        }
        else
        {
            UpdateMaxMinSim(similarityVec[i]);
        }
    }
    if (nSim < nHNode*nGNode)
    {
        minSim = 0;
    }
    return 0;
}

// this function reads the node similarity file that have following format:
//
// H_node_0 G_node_2
// H_node_1 G_node_1
// H_node_2 G_node_0
// ...
// H_node_i G_node_j
//
// H and G are two networks, each row represents an ortholog pair.

int SimilarityMatrix::InitializeBinarySimMatrix(Network& hNetwork, Network& gNetwork, const char* sFile)
{
    nHNode = hNetwork.nNode;
    nGNode = gNetwork.nNode;

    ifstream ifs(sFile);

    if (!ifs.good())
    {
        cerr << "Input file error!" << endl;
        exit(-1);
    }

    string hStrNode, gStrNode;
    unsigned int nCountTmp = 0;
    while(ifs >> hStrNode >> gStrNode)
    {
        simStrList[0].push_back(hStrNode);
        simStrList[1].push_back(gStrNode);
        hStrNode.clear();
        gStrNode.clear();
        nCountTmp += 1;
    }
    nSim = nCountTmp;

    if (nSim > nHNode*nGNode)
    {
        cerr << "Error: the number of orth pairs is larger than the maximal pair combinations ([#nodes in network1]*[#nodes in network2])." << endl;
        exit(0);
    }
    matrix = new float* [nHNode];

    for (int i = 0; i < nHNode; i++)
    {
        matrix[i] = new float[nGNode];
        for (int j = 0; j < nGNode; j++)
        {
            matrix[i][j] = 0;
        }
    }

    int hNode, gNode;
    for (int i = 0; i < nSim; i++)
    {
        hNode = hNetwork.GetNodeIndex(simStrList[0][i]);
        gNode = gNetwork.GetNodeIndex(simStrList[1][i]);
        matrix[hNode][gNode] = 1;
    }
    
    if (nSim != nHNode*nGNode)
    {
        minSim = 0;
    }
    if (nSim != 0)
    {
        maxSim = 1;
    }
    return 0;
}

int SimilarityMatrix::InitializeEmptySimMatrix(Network& hNetwork, Network& gNetwork)
{
    nHNode = hNetwork.nNode;
    nGNode = gNetwork.nNode;

    matrix = new float* [nHNode];

    for (int i = 0; i < nHNode; i++)
    {
        matrix[i] = new float[nGNode];
        for (int j = 0; j < nGNode; j++)
        {
            matrix[i][j] = 0;
        }
    }
    return 0;
}

int SimilarityMatrix::CopySimMatrix(SimilarityMatrix& simMatrix)
{
    nHNode = simMatrix.nHNode;
    nGNode = simMatrix.nGNode;
    nSim = simMatrix.nSim;
    maxSim = simMatrix.maxSim;
    minSim = simMatrix.minSim;

    matrix = new float* [nHNode];
    for (int i = 0; i < nHNode; i++)
    {
        matrix[i] = new float[nGNode];
        for (int j = 0; j < nGNode; j++)
        {
            matrix[i][j] = simMatrix.matrix[i][j];
        }
    }

    simStrList[0] = simMatrix.simStrList[0];
    simStrList[1] = simMatrix.simStrList[1];
    similarityVec = simMatrix.similarityVec;

    return 0;
}

int SimilarityMatrix::PrintMatrix()
{
    for (int i = 0; i < nHNode; i++)
    {
        for (int j = 0; j < nGNode; j++)
        {
            printf("%7.3f", matrix[i][j]);
            if (j != nGNode - 1) cout << " ";
        }
        cout << endl;
    }
    return 0;
}

int SimilarityMatrix::PrintSimilarityInfo()
{
    cout << "The No. of orth pairs: " << nSim << endl;
    return 0;
}

int SimilarityMatrix::MatrixNormalization()
{
    //cout << "maxSim : " << maxSim << endl;
    //cout << "minSim : " << minSim << endl;

    float maxAbsSim = abs(maxSim) > abs(minSim) ? abs(maxSim) : abs(minSim);

    for (int i = 0; i < nHNode; i++)
    {
        for (int j = 0; j < nGNode; j++)
        {
            matrix[i][j] /= maxAbsSim;
        }
    }
    maxSim /= maxAbsSim;
    minSim /= maxAbsSim;
    return 0;
}

int SimilarityMatrix::MatrixNormalization(float** matrix, float max, float min, float maxSimNeiSimMatrix, float minSimNeiSimMatrix)
{
    cout << "maxSim : " << max << endl;
    cout << "minSim : " << min << endl;
    cout << "maxSimNeiSimMatrix : " << maxSimNeiSimMatrix << endl;
    cout << "minSimNeiSimMatrix : " << minSimNeiSimMatrix << endl;
    float maxAbs = abs(max) > abs(min) ? abs(max) : abs(min);
    for (int i = 0; i < nHNode; i++)
    {
        for (int j = 0; j < nGNode; j++)
        {
            matrix[i][j] /= max;
        }
    }
    return 0;
}

int SimilarityMatrix::MatrixNormalization(float** matrix, float max)
{
    for (int i = 0; i < nHNode; i++)
    {
        for (int j = 0; j < nGNode; j++)
        {
            matrix[i][j] /= max;
        }
    }
    return 0;
}

int SimilarityMatrix::UpdateMaxMinSim(float sim)
{
    maxSim = maxSim > sim ? maxSim : sim;
    minSim = minSim < sim ? minSim : sim;
    return 0;
}

int SimilarityMatrix::GetEdgeDiffNMatrices(float ***edgeDiffNMatrix, int ***hNetNMatrix, int ***gNetNMatrix, Network& hNetwork, Network& gNetwork, int n)
{
    for (int i = 0; i < n; i++)
    {
        float maxEdgeDiff = 0;
        edgeDiffNMatrix[i] = new float *[nHNode];
        for (int j = 0; j < nHNode; j++)
        {
            edgeDiffNMatrix[i][j] = new float[nGNode];
            for (int k = 0; k < nGNode; k++)
            {
                vector<int> hNeighborIndex; //include the center vertex
                vector<int> gNeighborIndex; //include the center vertex
                hNeighborIndex.push_back(j);
                gNeighborIndex.push_back(k);
                for (int l = 0; l < nHNode; l++)
                {
                    if ((l != j) && (hNetNMatrix[i][j][l] != 0))
                    {
                        hNeighborIndex.push_back(l);
                    }
                }
                for (int l = 0; l < nGNode; l++)
                {
                    if ((l != k) && (gNetNMatrix[i][k][l] != 0))
                    {
                        gNeighborIndex.push_back(l);
                    }
                }


                int hNNeiEdge = 0;
                int gNNeiEdge = 0;
                for (int l = 0; l < hNeighborIndex.size()-1; l++)
                {
                    for (int m = l+1; m < hNeighborIndex.size(); m++)
                    {
                        hNNeiEdge += hNetwork.netMatrix[hNeighborIndex[l]][hNeighborIndex[m]];
                    }
                }
                for (int l = 0; l < gNeighborIndex.size()-1; l++)
                {
                    for (int m = l+1; m < gNeighborIndex.size(); m++)
                    {
                        gNNeiEdge += gNetwork.netMatrix[gNeighborIndex[l]][gNeighborIndex[m]];
                    }
                }
                edgeDiffNMatrix[i][j][k] = 1-fabs(hNNeiEdge-gNNeiEdge)/(hNNeiEdge+gNNeiEdge);
                maxEdgeDiff = maxEdgeDiff > edgeDiffNMatrix[i][j][k] ? maxEdgeDiff : edgeDiffNMatrix[i][j][k];
            }
        }
        //normalization
        for (int j = 0; j < nHNode; j++)
        {
            for (int k = 0; k < nGNode; k++)
            {
                edgeDiffNMatrix[i][j][k] /= maxEdgeDiff;
            }
        }
    }
    return 0;
}

int SimilarityMatrix::PrintEdgeDiffNMatrices(float ***edgeDiffNMatrix, int n)
{
    for (int i = 0; i < n; i++)
    {
        cout << "edgeDiffNMatrix : " << i << endl;
        for (int j = 0; j < nHNode; j++)
        {
            for (int k = 0; k < nGNode; k++)
            {
                printf("%6.4f", edgeDiffNMatrix[i][j][k]);
                if (k != nGNode - 1) cout << " ";
            }
            cout << endl;
        }
        cout << "-----------------------------------" << endl;
    }
    return 0;
}

int SimilarityMatrix::PrintSimDensNMatrices(float ***simDensNMatrix, int n)
{
    for (int i = 0; i < n; i++)
    {
        cout << "simDensNMatrix : " << i << endl;
        for (int j = 0; j < nHNode; j++)
        {
            for (int k = 0; k < nGNode; k++)
            {
                printf("%6.4f", simDensNMatrix[i][j][k]);
                if (k != nGNode - 1) cout << " ";
            }
            cout << endl;
        }
        cout << "-----------------------------------" << endl;
    }
    return 0;
}

int SimilarityMatrix::CleanEdgeDiffNMatrices(float ***edgeDiffNMatrix, int n)
{
    for (int i = 0; i < n ; i++)
    {
        for (int j = 0; j < nHNode; j++)
        {
            delete[] edgeDiffNMatrix[i][j];
            edgeDiffNMatrix[i][j] = NULL;
        }
        delete[] edgeDiffNMatrix[i];
        edgeDiffNMatrix[i] = NULL;
    }
    return 0;
}

int SimilarityMatrix::CleanSimDensNMatrices(float ***simDensNMatrix, int n)
{
    for (int i = 0; i < n ; i++)
    {
        for (int j = 0; j < nHNode; j++)
        {
            delete[] simDensNMatrix[i][j];
            simDensNMatrix[i][j] = NULL;
        }
        delete[] simDensNMatrix[i];
        simDensNMatrix[i] = NULL;
    }
    return 0;
}

int SimilarityMatrix::CleanFinalSimMatrix(int **finalSimMatrix)
{
    for (int i = 0; i < nHNode; i++)
    {
        delete[] finalSimMatrix[i];
        finalSimMatrix[i] = NULL;
    }
    delete[] finalSimMatrix;
    finalSimMatrix = NULL;

    return 0;
}

int SimilarityMatrix::GetSimGraphletMatrix(float**& simGraphletMatrix, Network& hNetwork, Network& gNetwork, float& maxSimGraphlet, float& minSimGraphlet)
{
    Graphlet *hGraphlets;
    Graphlet *gGraphlets;

    hGraphlets = new Graphlet [hNetwork.nLinkedNode]; //not including isolated nodes
    gGraphlets = new Graphlet [gNetwork.nLinkedNode];
    
    //cout << "  Generating graphlets..." << endl;
    for (int i = 0; i < hNetwork.nLinkedNode; i++)
    {
            hGraphlets[i].getGraphlet(hNetwork, hNetwork.linkedNodeIntVec[i]);
            //cout << "    network 1: " << i+1 << endl;
    }

    for (int i = 0; i < gNetwork.nLinkedNode; i++)
    {
            gGraphlets[i].getGraphlet(gNetwork, gNetwork.linkedNodeIntVec[i]);
            //cout << "    network 2: " << i+1 << endl;
    }

    int nPatternNodes[NPATTERN] = {3,3,3,4,4,4,4,4,4,4,4,4,4,4};
    int nPatternEdges[NPATTERN] = {2,2,3,3,3,3,3,4,4,4,4,5,5,6};

//    int patternIndex[] = {0,2,3,4,5,7,8,9,10,11,12,13};
//    int nPatternSelected = 13;
//    int patternIndex[] = {0,1,2,3,4,5,6,7,8,9,10,11,12,13};
//    int nPatternSelected = 14;

    int patternIndex[] = {0,3,4,5,7,8,9,11};
    int nPatternSelected = 8;

    simGraphletMatrix = new float *[hNetwork.nNode];
    
    for (int i = 0; i < hNetwork.nNode; i++)
    {
        simGraphletMatrix[i] = new float[gNetwork.nNode];
        for (int j = 0; j < gNetwork.nNode; j++)
        {
            cout << "subnetwork " << i+1 << " " << j+1 << endl;

            simGraphletMatrix[i][j] = 0;
            if (hNetwork.isNonIsolated[i] == false || gNetwork.isNonIsolated[j] == false)
            {
                continue;
            }

            // convert the network node index to linked node index
            int hLinkedNode = hNetwork.GetLinkedNodeIndex(i); 
            int gLinkedNode = gNetwork.GetLinkedNodeIndex(j);
            for (int k = 0; k < nPatternSelected; k++)
            {
                if (hGraphlets[hLinkedNode].vGraphletPattern[patternIndex[k]].size() <= gGraphlets[gLinkedNode].vGraphletPattern[patternIndex[k]].size())
                {
                    for (int l = 0; l < hGraphlets[hLinkedNode].vGraphletPattern[patternIndex[k]].size(); l++)
                    {
                        float maxScore = 0;
                        float score = 0;
                        for (int m = 0; m < gGraphlets[gLinkedNode].vGraphletPattern[patternIndex[k]].size(); m++)
                        {
                            int orthCount;
                            int loopCount;
                            if (patternIndex[k] == 1)
                            {
                                for (int p = 0; p < 2; p++)
                                {
                                    orthCount = 0;
                                    loopCount = 0;
                                    for (int n = 0; n < nPatternNodes[patternIndex[k]]; n++)
                                    {
                                        int node1 = hGraphlets[hLinkedNode].vGraphletPattern[patternIndex[k]][l][n];
                                        int node2;
                                        if (p == 1 && (n == 0 || n == 2))
                                        {
                                            node2 = gGraphlets[gLinkedNode].vGraphletPattern[patternIndex[k]][m][2-n]; //2,1,0
                                        }
                                        else
                                        {
                                            node2 = gGraphlets[gLinkedNode].vGraphletPattern[patternIndex[k]][m][n];
                                        }

                                        if ( matrix[hNetwork.linkedNodeIntVec[node1]][gNetwork.linkedNodeIntVec[node2]] > 0 )
                                        {
                                            orthCount += 1;
                                        }
                                        if (hNetwork.isLoop[hNetwork.linkedNodeIntVec[node1]] && gNetwork.isLoop[gNetwork.linkedNodeIntVec[node2]])
                                        {
                                            loopCount += 1;
                                        }
                                      }
                                    score = GetPatternSim(nPatternNodes[patternIndex[k]], nPatternEdges[patternIndex[k]], orthCount, loopCount);
                                    maxScore = maxScore > score ? maxScore : score;
                                }
                            }
                            else if (patternIndex[k] == 2)
                            {

                                for (int p = 0; p < 2; p++)
                                {
                                    orthCount = 0;
                                    loopCount = 0;
                                    for (int n = 0; n < nPatternNodes[patternIndex[k]]; n++)
                                    {
                                        int node1 = hGraphlets[hLinkedNode].vGraphletPattern[patternIndex[k]][l][n];
                                        int node2;
                                        if (p == 1 && (n == 1 || n == 2))
                                        {
                                            node2 = gGraphlets[gLinkedNode].vGraphletPattern[patternIndex[k]][m][3-n]; //0,2,1
                                        }
                                        else
                                        {
                                            node2 = gGraphlets[gLinkedNode].vGraphletPattern[patternIndex[k]][m][n];
                                        }

                                        if ( matrix[hNetwork.linkedNodeIntVec[node1]][gNetwork.linkedNodeIntVec[node2]] > 0 )
                                        {
                                            orthCount += 1;
                                        }
                                        if (hNetwork.isLoop[hNetwork.linkedNodeIntVec[node1]] && gNetwork.isLoop[gNetwork.linkedNodeIntVec[node2]])
                                        {
                                            loopCount += 1;
                                        }
                                    }
                                    score = GetPatternSim(nPatternNodes[patternIndex[k]], nPatternEdges[patternIndex[k]], orthCount, loopCount);
                                    maxScore = maxScore > score ? maxScore : score;
                                }
                            }
                            else if (patternIndex[k] == 10 ||
                                     patternIndex[k] == 8 ||
                                     patternIndex[k] == 5)
                            {

                                for (int p = 0; p < 2; p++)
                                {
                                    orthCount = 0;
                                    loopCount = 0;
                                    for (int n = 0; n < nPatternNodes[patternIndex[k]]; n++)
                                    {
                                        int node1 = hGraphlets[hLinkedNode].vGraphletPattern[patternIndex[k]][l][n];
                                        int node2;
                                        if (p == 1 && (n == 2 || n == 3))
                                        {
                                            node2 = gGraphlets[gLinkedNode].vGraphletPattern[patternIndex[k]][m][5-n]; //0,1,3,2
                                        }
                                        else
                                        {
                                            node2 = gGraphlets[gLinkedNode].vGraphletPattern[patternIndex[k]][m][n];
                                        }

                                        if ( matrix[hNetwork.linkedNodeIntVec[node1]][gNetwork.linkedNodeIntVec[node2]] > 0 )
                                        {
                                            orthCount += 1;
                                        }
                                        if (hNetwork.isLoop[hNetwork.linkedNodeIntVec[node1]] && gNetwork.isLoop[gNetwork.linkedNodeIntVec[node2]])
                                        {
                                            loopCount += 1;
                                        }
                                    }
                                    score = GetPatternSim(nPatternNodes[patternIndex[k]], nPatternEdges[patternIndex[k]], orthCount, loopCount);
                                    maxScore = maxScore > score ? maxScore : score;
                                }
                            }
                            else if (patternIndex[k] == 7 ||
                                     patternIndex[k] == 11 ||
                                     patternIndex[k] == 12)
                            {
                                for (int p = 0; p < 2; p++)
                                {
                                    orthCount = 0;
                                    loopCount = 0;
                                    for (int n = 0; n < nPatternNodes[patternIndex[k]]; n++)
                                    {
                                        int node1 = hGraphlets[hLinkedNode].vGraphletPattern[patternIndex[k]][l][n];
                                        int node2;
                                        if (p == 1 && (n == 1 || n == 3))
                                        {
                                            node2 = gGraphlets[gLinkedNode].vGraphletPattern[patternIndex[k]][m][4-n]; //0,3,2,1
                                        }
                                        else
                                        {
                                            node2 = gGraphlets[gLinkedNode].vGraphletPattern[patternIndex[k]][m][n];
                                        }

                                        if ( matrix[hNetwork.linkedNodeIntVec[node1]][gNetwork.linkedNodeIntVec[node2]] > 0 )
                                        {
                                            orthCount += 1;
                                        }
                                        if (hNetwork.isLoop[hNetwork.linkedNodeIntVec[node1]] && gNetwork.isLoop[gNetwork.linkedNodeIntVec[node2]])
                                        {
                                            loopCount += 1;
                                        }
                                    }
                                    score = GetPatternSim(nPatternNodes[patternIndex[k]], nPatternEdges[patternIndex[k]], orthCount, loopCount);
                                    maxScore = maxScore > score ? maxScore : score;
                                }
                            }
                            else if (patternIndex[k] == 6 ||
                                     patternIndex[k] == 13)
                            {
                                int permutation[6][3] = {{1,2,3},{1,3,2},{2,1,3},{2,3,1},{3,1,2},{3,2,1}};

                                for (int p = 0; p < 6; p++)
                                {
                                    orthCount = 0;
                                    loopCount = 0;
                                    //node0
                                    int node1 = hGraphlets[hLinkedNode].vGraphletPattern[patternIndex[k]][l][0];
                                    int node2 = gGraphlets[gLinkedNode].vGraphletPattern[patternIndex[k]][m][0];

                                    if ( matrix[hNetwork.linkedNodeIntVec[node1]][gNetwork.linkedNodeIntVec[node2]] > 0 )
                                    {
                                        orthCount += 1;
                                    }
                                    if (hNetwork.isLoop[hNetwork.linkedNodeIntVec[node1]] && gNetwork.isLoop[gNetwork.linkedNodeIntVec[node2]])
                                    {
                                        loopCount += 1;
                                    }
                                    for (int n = 1; n < nPatternNodes[patternIndex[k]]; n++)
                                    {
                                        node1 = hGraphlets[hLinkedNode].vGraphletPattern[patternIndex[k]][l][n];
                                        node2 = gGraphlets[gLinkedNode].vGraphletPattern[patternIndex[k]][m][permutation[p][n-1]];

                                        if ( matrix[hNetwork.linkedNodeIntVec[node1]][gNetwork.linkedNodeIntVec[node2]] > 0 )
                                        {
                                            orthCount += 1;
                                        }
                                        if (hNetwork.isLoop[hNetwork.linkedNodeIntVec[node1]] && gNetwork.isLoop[gNetwork.linkedNodeIntVec[node2]])
                                        {
                                            loopCount += 1;
                                        }
                                    }
                                    score = GetPatternSim(nPatternNodes[patternIndex[k]], nPatternEdges[patternIndex[k]], orthCount, loopCount);
                                    maxScore = maxScore > score ? maxScore : score;
                                }
                            }
                            else
                            {
                                orthCount = 0;
                                loopCount = 0;
                                for (int n = 0; n < nPatternNodes[patternIndex[k]]; n++)
                                {
                                    int node1 = hGraphlets[hLinkedNode].vGraphletPattern[patternIndex[k]][l][n];
                                    int node2 = gGraphlets[gLinkedNode].vGraphletPattern[patternIndex[k]][m][n];
                                    if ( matrix[hNetwork.linkedNodeIntVec[node1]][gNetwork.linkedNodeIntVec[node2]] > 0 )
                                    {
                                        orthCount += 1;
                                    }
                                    if (hNetwork.isLoop[hNetwork.linkedNodeIntVec[node1]] && gNetwork.isLoop[gNetwork.linkedNodeIntVec[node2]])
                                    {
                                        loopCount += 1;
                                    }
                                }
                                score = GetPatternSim(nPatternNodes[patternIndex[k]], nPatternEdges[patternIndex[k]], orthCount, loopCount);
                                maxScore = maxScore > score ? maxScore : score;
                            }
                        }
                        simGraphletMatrix[i][j] += maxScore;
                    }
                }
                else
                {
                    for (int m = 0; m < gGraphlets[gLinkedNode].vGraphletPattern[patternIndex[k]].size(); m++)
                    {
                        float maxScore = 0;
                        float score = 0;
                        for (int l = 0; l < hGraphlets[hLinkedNode].vGraphletPattern[patternIndex[k]].size(); l++)
                        {
                            int orthCount;
                            int loopCount;
                            if (patternIndex[k] == 1)
                            {
                                for (int p = 0; p < 2; p++)
                                {
                                    orthCount = 0;
                                    loopCount = 0;
                                    for (int n = 0; n < nPatternNodes[patternIndex[k]]; n++)
                                    {
                                        int node1 = hGraphlets[hLinkedNode].vGraphletPattern[patternIndex[k]][l][n];
                                        int node2;
                                        if (p == 1 && (n == 0 || n == 2))
                                        {
                                            node2 = gGraphlets[gLinkedNode].vGraphletPattern[patternIndex[k]][m][2-n]; //2,1,0
                                        }
                                        else
                                        {
                                            node2 = gGraphlets[gLinkedNode].vGraphletPattern[patternIndex[k]][m][n];
                                        }

                                        if ( matrix[hNetwork.linkedNodeIntVec[node1]][gNetwork.linkedNodeIntVec[node2]] > 0 )
                                        {
                                            orthCount += 1;
                                        }
                                        if (hNetwork.isLoop[hNetwork.linkedNodeIntVec[node1]] && gNetwork.isLoop[gNetwork.linkedNodeIntVec[node2]])
                                        {
                                            loopCount += 1;
                                        }
                                    }
                                    score = GetPatternSim(nPatternNodes[patternIndex[k]], nPatternEdges[patternIndex[k]], orthCount, loopCount);
                                    maxScore = maxScore > score ? maxScore : score;
                                }
                            }
                            else if (patternIndex[k] == 2)
                            {
                                for (int p = 0; p < 2; p++)
                                {
                                    orthCount = 0;
                                    loopCount = 0;
                                    for (int n = 0; n < nPatternNodes[patternIndex[k]]; n++)
                                    {
                                        int node1 = hGraphlets[hLinkedNode].vGraphletPattern[patternIndex[k]][l][n];
                                        int node2;
                                        if (p == 1 && (n == 1 || n == 2))
                                        {
                                            node2 = gGraphlets[gLinkedNode].vGraphletPattern[patternIndex[k]][m][3-n]; //0,2,1
                                        }
                                        else
                                        {
                                            node2 = gGraphlets[gLinkedNode].vGraphletPattern[patternIndex[k]][m][n];
                                        }

                                        if ( matrix[hNetwork.linkedNodeIntVec[node1]][gNetwork.linkedNodeIntVec[node2]] > 0 )
                                        {
                                            orthCount += 1;
                                        }
                                        if (hNetwork.isLoop[hNetwork.linkedNodeIntVec[node1]] && gNetwork.isLoop[gNetwork.linkedNodeIntVec[node2]])
                                        {
                                            loopCount += 1;
                                        }
                                    }
                                    score = GetPatternSim(nPatternNodes[patternIndex[k]], nPatternEdges[patternIndex[k]], orthCount, loopCount);
                                    maxScore = maxScore > score ? maxScore : score;
                                }
                            }
                            else if (patternIndex[k] == 10 || patternIndex[k] == 8 || patternIndex[k] == 5)
                            {
                                for (int p = 0; p < 2; p++)
                                {
                                    orthCount = 0;
                                    loopCount = 0;
                                    for (int n = 0; n < nPatternNodes[patternIndex[k]]; n++)
                                    {
                                        int node1 = hGraphlets[hLinkedNode].vGraphletPattern[patternIndex[k]][l][n];
                                        int node2;
                                        if (p == 1 && (n == 2 || n == 3))
                                        {
                                            node2 = gGraphlets[gLinkedNode].vGraphletPattern[patternIndex[k]][m][5-n]; //0,1,3,2
                                        }
                                        else
                                        {
                                            node2 = gGraphlets[gLinkedNode].vGraphletPattern[patternIndex[k]][m][n];
                                        }

                                        if ( matrix[hNetwork.linkedNodeIntVec[node1]][gNetwork.linkedNodeIntVec[node2]] > 0 )
                                        {
                                            orthCount += 1;
                                        }
                                        if (hNetwork.isLoop[hNetwork.linkedNodeIntVec[node1]] && gNetwork.isLoop[gNetwork.linkedNodeIntVec[node2]])
                                        {
                                            loopCount += 1;
                                        }
                                    }
                                    score = GetPatternSim(nPatternNodes[patternIndex[k]], nPatternEdges[patternIndex[k]], orthCount, loopCount);
                                    maxScore = maxScore > score ? maxScore : score;
                                }
                            }
                            else if (patternIndex[k] == 7 || patternIndex[k] == 11 || patternIndex[k] == 12)
                            {
                                for (int p = 0; p < 2; p++)
                                {
                                    orthCount = 0;
                                    loopCount = 0;
                                    for (int n = 0; n < nPatternNodes[patternIndex[k]]; n++)
                                    {
                                        int node1 = hGraphlets[hLinkedNode].vGraphletPattern[patternIndex[k]][l][n];
                                        int node2;
                                        if (p == 1 && (n == 1 || n == 3))
                                        {
                                            node2 = gGraphlets[gLinkedNode].vGraphletPattern[patternIndex[k]][m][4-n]; //0,3,2,1
                                        }
                                        else
                                        {
                                            node2 = gGraphlets[gLinkedNode].vGraphletPattern[patternIndex[k]][m][n];
                                        }

                                        if ( matrix[hNetwork.linkedNodeIntVec[node1]][gNetwork.linkedNodeIntVec[node2]] > 0 )
                                        {
                                            orthCount += 1;
                                        }
                                        if (hNetwork.isLoop[hNetwork.linkedNodeIntVec[node1]] && gNetwork.isLoop[gNetwork.linkedNodeIntVec[node2]])
                                        {
                                            loopCount += 1;
                                        }
                                    }
                                    score = GetPatternSim(nPatternNodes[patternIndex[k]], nPatternEdges[patternIndex[k]], orthCount, loopCount);
                                    maxScore = maxScore > score ? maxScore : score;
                                }
                            }
                            else if (patternIndex[k] == 6 || patternIndex[k] == 13)
                            {
                                int permutation[6][3] = {{1,2,3},{1,3,2},{2,1,3},{2,3,1},{3,1,2},{3,2,1}};

                                for (int p = 0; p < 6; p++)
                                {
                                    orthCount = 0;
                                    loopCount = 0;
                                    //node0
                                    int node1 = hGraphlets[hLinkedNode].vGraphletPattern[patternIndex[k]][l][0];
                                    int node2 = gGraphlets[gLinkedNode].vGraphletPattern[patternIndex[k]][m][0];

                                    if ( matrix[hNetwork.linkedNodeIntVec[node1]][gNetwork.linkedNodeIntVec[node2]] > 0 )
                                    {
                                        orthCount += 1;
                                    }
                                    if (hNetwork.isLoop[hNetwork.linkedNodeIntVec[node1]] && gNetwork.isLoop[gNetwork.linkedNodeIntVec[node2]])
                                    {
                                        loopCount += 1;
                                    }
                                    for (int n = 1; n < nPatternNodes[patternIndex[k]]; n++)
                                    {
                                        node1 = hGraphlets[hLinkedNode].vGraphletPattern[patternIndex[k]][l][n];
                                        node2 = gGraphlets[gLinkedNode].vGraphletPattern[patternIndex[k]][m][permutation[p][n-1]];

                                        if ( matrix[hNetwork.linkedNodeIntVec[node1]][gNetwork.linkedNodeIntVec[node2]] > 0 )
                                        {
                                            orthCount += 1;
                                        }
                                        if (hNetwork.isLoop[hNetwork.linkedNodeIntVec[node1]] && gNetwork.isLoop[gNetwork.linkedNodeIntVec[node2]])
                                        {
                                            loopCount += 1;
                                        }
                                    }
                                    score = GetPatternSim(nPatternNodes[patternIndex[k]], nPatternEdges[patternIndex[k]], orthCount, loopCount);
                                    maxScore = maxScore > score ? maxScore : score;
                                }
                            }
                            else
                            {
                                orthCount = 0;
                                loopCount = 0;
                                for (int n = 0; n < nPatternNodes[patternIndex[k]]; n++)
                                {

                                    int node1 = hGraphlets[hLinkedNode].vGraphletPattern[patternIndex[k]][l][n];
                                    int node2 = gGraphlets[gLinkedNode].vGraphletPattern[patternIndex[k]][m][n];

                                    if ( matrix[hNetwork.linkedNodeIntVec[node1]][gNetwork.linkedNodeIntVec[node2]] > 0 )
                                    {
                                        orthCount += 1;
                                    }
                                    if (hNetwork.isLoop[hNetwork.linkedNodeIntVec[node1]] && gNetwork.isLoop[gNetwork.linkedNodeIntVec[node2]])
                                    {
                                        loopCount += 1;
                                    }
                                }
                                score = GetPatternSim(nPatternNodes[patternIndex[k]], nPatternEdges[patternIndex[k]], orthCount, loopCount);
                                maxScore = maxScore > score ? maxScore : score;
                            }
                        }
                        simGraphletMatrix[i][j] += maxScore;
                    }
                }
            }
            if (i == 0 && j == 0)
            {
                maxSimGraphlet = simGraphletMatrix[i][j];
                minSimGraphlet = simGraphletMatrix[i][j];
            }
            else
            {
                maxSimGraphlet = simGraphletMatrix[i][j] > maxSimGraphlet ? simGraphletMatrix[i][j] : maxSimGraphlet;
                minSimGraphlet = simGraphletMatrix[i][j] < minSimGraphlet ? simGraphletMatrix[i][j] : minSimGraphlet;
            }
        }
    }


    return 0;
}

float SimilarityMatrix::GetPatternSim(int nNode, int nEdge, int orthCount, int loopCount)
{
    float similarity = NODE_WEIGHT*ORTH_SCORE*orthCount + EDGE_WEIGHT*EDGE_SCORE*(nEdge+loopCount);
    return similarity;
}

int SimilarityMatrix::PrintSimGraphletMatrix(int** simGraphletMatrix)
{
    for (int i = 0; i < nHNode; i++)
    {
        for (int j = 0; j < nGNode; j++)
        {
            printf("%7d ", simGraphletMatrix[i][j]);
        }
        cout << endl;
    }
    return 0;
}

int SimilarityMatrix::PrintSimGraphletMatrix(float** simGraphletMatrix)
{
    for (int i = 0; i < nHNode; i++)
    {
        for (int j = 0; j < nGNode; j++)
        {
            printf("%7.2f ", simGraphletMatrix[i][j]);
        }
        cout << endl;
    }
    return 0;
}
int SimilarityMatrix::PrintUnalignedNodePairs()
{
    for (int i = 0; i < unalignedNodePairs[0].size(); i++)
    {
        cout << unalignedNodePairs[0][i] << " " << unalignedNodePairs[1][i] << endl;
    }
    return 0;
}

int SimilarityMatrix::CleanSimGraphletMatrix(float** simGraphletMatrix)
{
    for (int i = 0; i < nHNode ; i++)
    {
        delete[] simGraphletMatrix[i];
        simGraphletMatrix[i] = NULL;
    }
    delete simGraphletMatrix;

    return 0;
}

